@@ -19,45 +19,32 @@
 #define POWER_PROFILES_DBUS_PATH          "/net/hadess/PowerProfiles"
 #define POWER_PROFILES_IFACE_NAME         POWER_PROFILES_DBUS_NAME
 
-GMainLoop *main_loop = NULL;
-
 typedef struct {
+  GMainLoop *main_loop;
   GDBusNodeInfo *introspection_data;
   GDBusConnection *connection;
   guint name_id;
+  gboolean was_started;
   int ret;
 
   PpdProfile active_profile;
-  GPtrArray *drivers;
+  GPtrArray *probed_drivers;
+  PpdDriver *driver;
   GPtrArray *actions;
 } PpdApp;
 
-static PpdDriver *
-get_driver_for_profile (PpdApp     *data,
-                        PpdProfile  profile)
-{
-  guint i;
+static PpdApp *ppd_app = NULL;
 
-  g_return_val_if_fail (ppd_profile_has_single_flag (profile), NULL);
+static void stop_profile_drivers (PpdApp *data);
+static void start_profile_drivers (PpdApp *data);
 
-  for (i = 0; i < data->drivers->len; i++) {
-    PpdDriver *driver = g_ptr_array_index (data->drivers, i);
-
-    if (ppd_driver_get_profiles (driver) & profile)
-      return driver;
-  }
-
-  return NULL;
-}
-
-#define GET_DRIVER(p) (get_driver_for_profile (data, p))
-#define ACTIVE_DRIVER (get_driver_for_profile (data, data->active_profile))
+#define GET_DRIVER(p) (ppd_driver_get_profiles (data->driver) & p ? data->driver : NULL)
+#define ACTIVE_DRIVER (data->driver)
 
 /* profile drivers and actions */
 #include "ppd-action-trickle-charge.h"
-#include "ppd-driver-balanced.h"
-#include "ppd-driver-power-saver.h"
-#include "ppd-driver-lenovo-dytc.h"
+#include "ppd-driver-placeholder.h"
+#include "ppd-driver-platform-profile.h"
 #include "ppd-driver-intel-pstate.h"
 #include "ppd-driver-fake.h"
 
@@ -66,12 +53,11 @@ typedef GType (*GTypeGetFunc) (void);
 static GTypeGetFunc objects[] = {
   /* Hardware specific profile drivers */
   ppd_driver_fake_get_type,
-  ppd_driver_lenovo_dytc_get_type,
+  ppd_driver_platform_profile_get_type,
   ppd_driver_intel_pstate_get_type,
 
-  /* Generic profile drivers */
-  ppd_driver_balanced_get_type,
-  ppd_driver_power_saver_get_type,
+  /* Generic profile driver */
+  ppd_driver_placeholder_get_type,
 
   /* Actions */
   ppd_action_trickle_charge_get_type,
@@ -95,8 +81,8 @@ get_active_profile (PpdApp *data)
 static const char *
 get_performance_inhibited (PpdApp *data)
 {
-  PpdDriver *driver;
   const char *ret;
+  PpdDriver *driver;
 
   driver = GET_DRIVER(PPD_PROFILE_PERFORMANCE);
   if (!driver)
@@ -220,24 +206,21 @@ actions_activate_profile (GPtrArray *actions,
 }
 
 static void
-activate_target_profile (PpdApp     *data,
-                         PpdProfile  target_profile)
+activate_target_profile (PpdApp                     *data,
+                         PpdProfile                  target_profile,
+                         PpdProfileActivationReason  reason)
 {
-  guint i;
+  g_autoptr(GError) error = NULL;
 
-  g_debug ("Setting active profile '%s' (current: '%s')",
+  g_debug ("Setting active profile '%s' for reason '%s' (current: '%s')",
            ppd_profile_to_str (target_profile),
+           ppd_profile_activation_reason_to_str (reason),
            ppd_profile_to_str (data->active_profile));
 
-  for (i = 0; i < data->drivers->len; i++) {
-    PpdDriver *driver = g_ptr_array_index (data->drivers, i);
-    g_autoptr(GError) error = NULL;
-
-    if (!ppd_driver_activate_profile (driver, target_profile, &error)) {
-      g_warning ("Failed to activate driver '%s': %s",
-                 ppd_driver_get_driver_name (driver),
-                 error->message);
-    }
+  if (!ppd_driver_activate_profile (data->driver, target_profile, reason, &error)) {
+    g_warning ("Failed to activate driver '%s': %s",
+               ppd_driver_get_driver_name (data->driver),
+               error->message);
   }
 
   actions_activate_profile (data->actions, target_profile);
@@ -259,11 +242,8 @@ set_active_profile (PpdApp      *data,
     return FALSE;
   }
 
-  if (target_profile == data->active_profile) {
-    g_set_error (error, G_DBUS_ERROR, G_DBUS_ERROR_FAILED,
-                 "Profile '%s' already active", profile);
-    return FALSE;
-  }
+  if (target_profile == data->active_profile)
+    return TRUE;
 
   if (target_profile == PPD_PROFILE_PERFORMANCE &&
       ppd_driver_is_performance_inhibited (GET_DRIVER (PPD_PROFILE_PERFORMANCE))) {
@@ -272,11 +252,10 @@ set_active_profile (PpdApp      *data,
     return FALSE;
   }
 
-  g_debug ("Transitioning active profile from '%s' to '%s'",
+  g_debug ("Transitioning active profile from '%s' to '%s' by user request",
            ppd_profile_to_str (data->active_profile), profile);
-  data->active_profile = target_profile;
 
-  activate_target_profile (data, target_profile);
+  activate_target_profile (data, target_profile, PPD_PROFILE_ACTIVATION_REASON_USER);
   send_dbus_event (data, PROP_ACTIVE_PROFILE);
 
   return TRUE;
@@ -307,7 +286,25 @@ driver_performance_inhibited_changed_cb (GObject    *gobject,
   if (!ppd_driver_is_performance_inhibited (driver))
     return;
 
-  activate_target_profile (data, PPD_PROFILE_BALANCED);
+  activate_target_profile (data, PPD_PROFILE_BALANCED, PPD_PROFILE_ACTIVATION_REASON_INHIBITION);
+  send_dbus_event (data, PROP_ACTIVE_PROFILE);
+}
+
+static void
+driver_profile_changed_cb (PpdDriver *driver,
+                           PpdProfile new_profile,
+                           gpointer   user_data)
+{
+  PpdApp *data = user_data;
+
+  g_debug ("Driver '%s' switched internally to profile '%s' (current: '%s')",
+           ppd_driver_get_driver_name (driver),
+           ppd_profile_to_str (new_profile),
+           ppd_profile_to_str (data->active_profile));
+  if (new_profile == data->active_profile)
+    return;
+
+  activate_target_profile (data, new_profile, PPD_PROFILE_ACTIVATION_REASON_INTERNAL);
   send_dbus_event (data, PROP_ACTIVE_PROFILE);
 }
 
@@ -372,8 +369,11 @@ name_lost_handler (GDBusConnection *connection,
                    const gchar     *name,
                    gpointer         user_data)
 {
+  PpdApp *data = user_data;
   g_debug ("power-profiles-daemon is already running, or it cannot own its D-Bus name. Verify installation.");
-  exit (0);
+  if (!data->was_started)
+    data->ret = 1;
+  g_main_loop_quit (data->main_loop);
 }
 
 static void
@@ -409,39 +409,31 @@ has_required_drivers (PpdApp *data)
   return TRUE;
 }
 
-static gboolean
-profile_already_handled (PpdApp     *data,
-                         PpdDriver  *driver,
-                         PpdProfile  profiles)
+static void
+driver_probe_request_cb (PpdDriver *driver,
+                         gpointer   user_data)
 {
-  guint i;
+  PpdApp *data = user_data;
 
-  for (i = 0; i < NUM_PROFILES; i++) {
-    PpdDriver *existing_driver;
-
-    if (!(profiles & (1 << i)))
-      continue;
-
-    existing_driver = GET_DRIVER(1 << i);
-    if (existing_driver) {
-      g_debug ("Driver '%s' conflicts with already probed driver '%s' for profile %s",
-               ppd_driver_get_driver_name (driver),
-               ppd_driver_get_driver_name (existing_driver),
-               ppd_profile_to_str (1 << i));
-      return TRUE;
-    }
-  }
-
-  return FALSE;
+  stop_profile_drivers (data);
+  start_profile_drivers (data);
 }
 
 static void
-name_acquired_handler (GDBusConnection *connection,
-                       const gchar     *name,
-                       gpointer         user_data)
+stop_profile_drivers (PpdApp *data)
+{
+  g_ptr_array_set_size (data->probed_drivers, 0);
+  g_ptr_array_set_size (data->actions, 0);
+  g_clear_object (&data->driver);
+}
+
+static void
+start_profile_drivers (PpdApp *data)
 {
-  PpdApp *data = user_data;
   guint i;
+  PpdProfile prev_profile;
+
+  prev_profile = data->active_profile;
 
   for (i = 0; i < G_N_ELEMENTS (objects); i++) {
     GObject *object;
@@ -450,9 +442,17 @@ name_acquired_handler (GDBusConnection *connection,
     if (PPD_IS_DRIVER (object)) {
       PpdDriver *driver = PPD_DRIVER (object);
       PpdProfile profiles;
+      PpdProbeResult result;
 
       g_debug ("Handling driver '%s'", ppd_driver_get_driver_name (driver));
 
+      if (data->driver != NULL) {
+        g_debug ("Driver '%s' already probed, skipping driver '%s'",
+                 ppd_driver_get_driver_name (data->driver),
+                 ppd_driver_get_driver_name (driver));
+        continue;
+      }
+
       profiles = ppd_driver_get_profiles (driver);
       if (!(profiles & PPD_PROFILE_ALL)) {
         g_warning ("Profile Driver '%s' implements invalid profiles '0x%X'",
@@ -462,22 +462,25 @@ name_acquired_handler (GDBusConnection *connection,
         continue;
       }
 
-      if (profile_already_handled (data, driver, profiles)) {
-        g_object_unref (object);
-        continue;
-      }
-
-      if (!ppd_driver_probe (driver)) {
+      result = ppd_driver_probe (driver, &prev_profile);
+      if (result == PPD_PROBE_RESULT_FAIL) {
         g_debug ("probe() failed for driver %s, skipping",
                  ppd_driver_get_driver_name (driver));
         g_object_unref (object);
         continue;
+      } else if (result == PPD_PROBE_RESULT_DEFER) {
+        g_signal_connect (G_OBJECT (driver), "probe-request",
+                          G_CALLBACK (driver_probe_request_cb), data);
+        g_ptr_array_add (data->probed_drivers, driver);
+        continue;
       }
 
-      g_ptr_array_add (data->drivers, driver);
+      data->driver = driver;
 
       g_signal_connect (G_OBJECT (driver), "notify::performance-inhibited",
                         G_CALLBACK (driver_performance_inhibited_changed_cb), data);
+      g_signal_connect (G_OBJECT (driver), "profile-changed",
+                        G_CALLBACK (driver_profile_changed_cb), data);
     } else if (PPD_IS_ACTION (object)) {
       PpdAction *action = PPD_ACTION (object);
 
@@ -501,17 +504,42 @@ name_acquired_handler (GDBusConnection *connection,
     goto bail;
   }
 
+  if (prev_profile != data->active_profile) {
+    g_debug ("Using '%s' as current profile from probed driver",
+             ppd_profile_to_str (prev_profile));
+    data->active_profile = prev_profile;
+  }
+
   /* Set initial state */
-  activate_target_profile (data, data->active_profile);
+  activate_target_profile (data, data->active_profile, PPD_PROFILE_ACTIVATION_REASON_RESET);
 
   send_dbus_event (data, PROP_ALL);
 
+  data->was_started = TRUE;
+
   return;
 
 bail:
-  data->ret = 0;
+  data->ret = 1;
   g_debug ("Exiting because some non recoverable error occurred during startup");
-  g_main_loop_quit (main_loop);
+  g_main_loop_quit (data->main_loop);
+}
+
+void
+restart_profile_drivers (void)
+{
+  stop_profile_drivers (ppd_app);
+  start_profile_drivers (ppd_app);
+}
+
+static void
+name_acquired_handler (GDBusConnection *connection,
+                       const gchar     *name,
+                       gpointer         user_data)
+{
+  PpdApp *data = user_data;
+
+  start_profile_drivers (data);
 }
 
 static gboolean
@@ -555,13 +583,21 @@ free_app_data (PpdApp *data)
     data->name_id = 0;
   }
 
+  g_ptr_array_free (data->probed_drivers, TRUE);
   g_ptr_array_free (data->actions, TRUE);
-  g_ptr_array_free (data->drivers, TRUE);
+  g_clear_object (&data->driver);
 
+  g_clear_pointer (&data->main_loop, g_main_loop_unref);
   g_clear_pointer (&data->introspection_data, g_dbus_node_info_unref);
   g_clear_object (&data->connection);
-  g_clear_pointer (&main_loop, g_main_loop_unref);
   g_free (data);
+  ppd_app = NULL;
+}
+
+void
+main_loop_quit (void)
+{
+  g_main_loop_quit (ppd_app->main_loop);
 }
 
 int main (int argc, char **argv)
@@ -592,15 +628,16 @@ int main (int argc, char **argv)
     g_setenv ("G_MESSAGES_DEBUG", "all", TRUE);
 
   data = g_new0 (PpdApp, 1);
+  data->main_loop = g_main_loop_new (NULL, TRUE);
+  data->probed_drivers = g_ptr_array_new_with_free_func ((GDestroyNotify) g_object_unref);
   data->actions = g_ptr_array_new_with_free_func ((GDestroyNotify) g_object_unref);
-  data->drivers = g_ptr_array_new_with_free_func ((GDestroyNotify) g_object_unref);
   data->active_profile = PPD_PROFILE_BALANCED;
+  ppd_app = data;
 
   /* Set up D-Bus */
   setup_dbus (data, replace);
 
-  main_loop = g_main_loop_new (NULL, TRUE);
-  g_main_loop_run (main_loop);
+  g_main_loop_run (data->main_loop);
   ret = data->ret;
   free_app_data (data);
 
